; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real ArcTan                                     *
; *                                                     *
; *     Copyright (C) 1989-1994 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   F48FATN

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Externals

             EXTRN   RealAdd     : NEAR
             EXTRN   RealPoly    : NEAR
             EXTRN   RealTrunc   : NEAR
             EXTRN   RealFloat   : NEAR
             EXTRN   RealDivRev  : NEAR
             EXTRN   RealMulNChk2: NEAR
             EXTRN   CmpAbsValue : NEAR
             EXTRN   ShortMulRev : NEAR

; Publics

             PUBLIC  RArcTan

;-------------------------------------------------------------------------------
; RArcTan computes the arctangent of a TURBO-Pascal six byte floating-point
; number. No overflow is possible, since function values range from -pi/2 to
; pi/2. The argument is reduced such that |z| <= 1/16 by an interval scheme
; based on the identity Arctan (x) = Arctan (a) + Arctan ((x-a) / (1 + x*a)).
; The arctangent of the reduced argument is then computed using a fast poly-
; nomial approximation. The result is finally adjusted depending on which
; reduction interval the argument was in. This polynomial approximation to the
; arctangent is used:
;
; p(z):=((-1.420288085938e-1*x^2+1.999981270101e-1)*x^2
;         -3.333333321307e-1)*x^2*x+x
;
; This approximation has a theoretical maximum relative error of 2.98e-13.
; Maximum observed error when evaluated in REAL arithmetic is 1.11e-12.
;
; INPUT:     DX:BX:AX  argument
;
; OUTPUT:    DX:BX:AX  arctan of argument
;
; DESTROYS:  AX,BX,CX,DX,SI,DI,Flags
;-------------------------------------------------------------------------------

RArcTan      PROC    FAR

             cmp     al, 06dh          ; x very small ?
             jb      $atn_x            ; yes, return x

             push    bp                ; save TURBO-Pascal frame pointer
             mov     ch, 80h           ; load mask for sign bit
             and     ch, dh            ; isolate sign bit
             xor     dh, ch            ; absolute value of x
             push    cx                ; save sign bit
             mov     cx, 81h           ; load
             xor     si, si            ;  floating
             xor     di, di            ;   constant 1
             call    CmpAbsValue       ; x <= 1 ?

             pushf                     ; save division flag
             jb      $no_division      ; x <= 1, no division

             call    RealDivRev        ; compute x = 1/x

$no_division:mov     di, dx            ; save
             mov     si, bx            ;  x
             push    ax                ;
             add     al, 3             ; x * 8
             mov     ch, 0ffh          ; signal rounding desired
             call    RealTrunc         ; n = Round (x * 8)
             mov     bp, ax            ; save n
             pop     cx                ; DI:SI:CX = x
             test    ax, ax            ; n = 0 ?
             mov     dx, di            ; reload x
             mov     bx, si            ;  into
             xchg    ax, cx            ;   DX:BX:AX
             jz      $atn_appr         ; n=0, evaluate approximating polynomial

             xchg    ax, cx            ; AX=n, CX = LSW of x
             push    ax                ; save n
             push    cx                ; save LSW of x
             cwd                       ; convert n to longint (n <= 8)
             call    RealFloat         ; convert n to REAL

             pop     cx                ; get LSW of x
             sub     al, 3             ; n/8
             mov     bp, ax            ; save
             mov     ah, dh            ; AX = packed n
             xchg    ax, bp            ; BP = packed n
             xor     dh, 80h           ; -n/8
             push    di                ; save
             push    si                ;  x
             push    cx                ; save x
             call    RealAdd           ; x - n/8

             pop     cx                ; get
             pop     si                ;  x
             pop     di                ;   from stack
             push    dx                ; save x - n/8
             push    bx                ;  on
             push    ax                ;   stack
             mov     ax, bp            ; get packed floating point n
             xor     dx, dx            ; unpack
             xor     bx, bx            ;  floating point
             xchg    ah, dh            ;   n
             call    ShortMulRev       ; x * n

             mov     cx, 81h           ; load
             xor     si, si            ;  real constant
             xor     di, di            ;   1.0
             call    RealAdd           ; compute 1 + x*n/8

             pop     cx                ; get
             pop     si                ;  back
             pop     di                ;   x - n/8
             call    RealDivRev        ; compute (x - n/8) / (1 + x*n/8)

             pop     bp                ; get integer n
             mov     cx, bp            ; compute
             add     cx, cx            ;  offset into
             add     bp, cx            ;   constant field
             add     bp, bp            ;    as 6*n
             add     bp, offset ATN_CST; address of constant B for interval n

$atn_appr:   mov     cx, 3             ; approximation uses three coefficients
             xor     si, si            ; polynomial of type P(x)*x+x
             mov     di, offset AT_COEFF; pointer to first coefficient
             call    RealPoly          ; compute arctan (z) = z + z * P(z)

             cmp     bp, offset ATN_CST + 6; first or second interval ?
             jb      $first_intv       ; first interval, no correction needed

             jne     $normal_add       ; not second interval, no special add

             mov     cx, 03365h        ; add eps for second interval
             mov     si, 07b6eh        ;  for a pseudo
             mov     di, 05561h        ;   multiple precision
             call    RealAdd           ;    addition

$normal_add: mov     cx, cs:[bp - 6]   ; load value
             mov     si, cs:[bp - 4]   ;  of constant B appropriate
             mov     di, cs:[bp - 2]   ;   for reduction interval
             call    RealAdd           ; add B to arctan (z)

$first_intv: popf                      ; get flag for division
             jb      $no_add           ; if no division then done

             mov     cx, 02181h        ; load
             mov     si, 0daa2h        ;  real constant
             mov     di, 0490fh        ;   0.5*pi
             xor     dh, 80h           ; - arctan (z)
             call    RealAdd           ; compute 0.5*pi - arctan (z)

$no_add:     pop     cx                ; get sign mask
             pop     bp                ; restore TURBO-Pascal frame pointer
             or      dh, ch            ; make sign bit

$atn_x:      ret

ATN_CST      db      07Dh,000h,000h,0D4h,0ADh,07Eh ; arctan (1/8) - eps
             db      07Eh,064h,0C9h,0AFh,0DBh,07Ah ; arctan (2/8)
             db      07Fh,027h,00Fh,0CAh,0B0h,037h ; arctan (3/8)
             db      07Fh,00Eh,02Bh,038h,063h,06Dh ; arctan (4/8)
             db      080h,0F8h,05Eh,05Dh,000h,00Fh ; arctan (5/8)
             db      080h,035h,019h,07Dh,0BCh,024h ; arctan (6/8)
             db      080h,0C2h,02Bh,03Eh,005h,038h ; arctan (7/8)
             db      080h,021h,0A2h,0DAh,00Fh,049h ; arctan (8/8)

AT_COEFF     db      07Eh,               070h,091h ; -1.420288085938e-1
             db      07Eh,014h,01Bh,04Fh,0CCh,04Ch ;  1.999981270101e-1
             db      07Fh,056h,0A0h,0AAh,0AAh,0AAh ; -3.333333321307e-1

RArcTan      ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END