; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Real Round/Trunc                                *
; *                                                     *
; *     Copyright (C) 1989-1993 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   FP48RND

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Externals

             EXTRN   HaltError: NEAR

; Publics

             PUBLIC  RealTrunc
             PUBLIC  RTrunc
             PUBLIC  RRound

;-------------------------------------------------------------------------------
; RealTrunc converts a TURBO-Pascal six byte floatingpoint number to a four
; byte signed integer. Truncation or rounding can be requested by the caller
; by setting a flag. If the conversion results in a long integer overflow, the
; routine returns with the carry flag set. When rounding is selected, the
; routine complies with the IEEE "round to nearest or even" mode. For example,
; Round (4.5) = 4, but Round (5.5) = 6. Special care is taken to accomodate
; correct handling of the smallest LONGINT number 8000000h.
;
; INPUT:     DX:BX:AX  floating point number
;            CH        rounding flag ( 0 = trunc, all others = round)
;
; OUTPUT:    DX:AX     converted longint number
;            CF        set if overflow occured
;
; DESTROYS:  AX,BX,CX,DX,Flags
;-------------------------------------------------------------------------------

$long_zero:  xor     ax, ax            ; load
             cwd                       ;  zero into DX:AX
             retn                      ; exit

$too_big:    jnz     $ovrfl_err2       ; abs (number) > 2^32

             cmp     dh, 80h           ; num negative && abs (num) < 2^32-2^24 ?
             jne     $ovrfl_err2       ; no, overflow

             xor     al, al            ; clear sticky flag
             push    dx                ; save original sign
             or      dh, 80h           ; set hidden bit
             jmp     $shft_done        ; too big numbers caught by 2nd check

$ovrfl_err2: stc                       ; signal error
             retn                      ; exit

             INCLUDE ALIGN.ASM

RealTrunc    PROC    NEAR

             add     al, 60h           ; number to big ?
             jc      $too_big          ; probably, do detailed check

             cmp     al, 0E0h          ; number < 0.5 ?
             jb      $long_zero        ; return zero

$size_ok:    push    dx                ; save sign
             or      dh, 80h           ; set implicit mantissa bit
             mov     cl, al            ; counter
             xor     al, al            ; initialize sticky flag
             cmp     cl, -16           ; 16-bit shift possible ?
             ja      $byte_shift       ; no, try 8-bit shift

             or      al, ah            ; accumulate
             or      al, bl            ;  sticky flag
             mov     ah, bh            ; shift DX:BX:AH
             mov     bx, dx            ;  16 bits to
             xor     dx, dx            ;   the right
             add     cl, 16            ; remaining bit shifts
             jz      $shft_done        ; no shifts left, ->

$byte_shift: cmp     cl, -8            ; 8-bit shift possible ?
             ja      $4bit_shift       ; no, try nibble shift

             or      al, ah            ; accumulate sticky flag
             mov     ah, bl            ; shift
             mov     bl, bh            ;  DX:BX:AH
             mov     bh, dl            ;   8 bits to
             movzx   dx, dh            ;    the right
             add     cl, 8             ; remaining bit shifts
             jz      $shft_done        ; no bit shifts left

$4bit_shift: neg     al                ; sticky flag <> 0 ?
             sbb     al, al            ; set to FFh if not 0
             cmp     cl, -4            ; nibble shift possible ?
             ja      $bit_shift        ; no, try single bit shifts

             shrd    ax, bx, 4         ; shift DX:BX:AH
             shrd    bx, dx, 4         ;  4s bit to
             shr     dx, 4             ;   the right and accumulate sticky flag
             add     cl, 4             ; remaining bit shifts
             jz      $shft_done        ; no shifts left

$bit_shift:  neg     al                ; sticky flag <> 0 ?
             sbb     al, al            ; set to FFh if not 0

             INCLUDE ALIGN.ASM

$shift_loop: shr     dx, 1             ; shift DX:BX:AH
             rcr     bx, 1             ;  1 bit to
             rcr     ax, 1             ;   the right and accumulate sticky flag
             inc     cl                ; adjust shift counter
             jnz     $shift_loop       ; until counter zero

$shft_done:  neg     ch                ; test if rounding flag set
             sbb     ch, ch            ; CH = FFh if rounding, CH = 0 if trunc
             and     ah, ch            ; clear fraction part if trunc
             add     ax, 8000h         ; round up ? AH = guard, AL = sticky
             jnz     $round            ; if no tie case (AH = 80, AL = 0)

             ror     bl, 1             ; move least significant
             rol     bl, 1             ;  bit into carry

$round:      pop     cx                ; get original sign flag
             adc     bx, 0             ; round up
             adc     dx, 0             ;  result if carry set
             xchg    ax, bx            ; result in DX:AX
             test    ch, ch            ; original argument negative ?
             jns     $pos_long         ; no, was positive

             not     dx                ; negate
             neg     ax                ;  longint
             sbb     dx, -1            ;   in DX:AX
             jnc     $rnd_done         ; DX:AX = 0, no need to check for ovrfl.

$pos_long:   xor     ch, dh            ; XOR sign of argument and sign of result
             add     ch, ch            ; CY, if signs differ (= overflow)

$rnd_done:   ret

RealTrunc    ENDP

             INCLUDE ALIGN.ASM

RTrunc       PROC    FAR

             xor     ch, ch            ; flag truncation
             call    RealTrunc         ; convert real to longint

             jc      RRangeError       ; longint overflowed

             ret

RTrunc       ENDP

             INCLUDE ALIGN.ASM

RRound       PROC    FAR

             mov     ch, 1             ; flag rounding
             call    RealTrunc         ; convert real to longint

             jc      RRangeError       ; longint overflowed

             ret

RRound       ENDP

RRangeError: mov     ax, 207           ; error code 207 (invalid fp operation)
             jmp     HaltError         ; execute error handler

             INCLUDE ALIGN.ASM

CODE         ENDS

             END