; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     Add Range to Set Routine                        *
; *                                                     *
; *     Copyright (C) 1991-1994 Norbert Juffa           *
; *     Copyright (C) 2000-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   SETHADR

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Publics

             PUBLIC  ZAddRange

;-------------------------------------------------------------------------------
; ZAddRange adds all elements between a specified low element number and a
; specified high element number to the set in the buffer. A mask is calculated
; to set the bit corresponding to the lowest element number, then the mask is
; rotated to generate the mask for the next higher bits, incrementing the buffer
; pointer when encountering byte boundaries, until all bits up to the one for
; the higher element number are set.
;
; On entry:  [SP+4]  Number of last element to add
;            [SP+6]  Number of first element to add
;            [SP+8]  Pointer to set buffer
;
; On exit:   [SP]    Pointer to set buffer
;-------------------------------------------------------------------------------

ZAddRange    PROC    FAR

             mov     bx, sp            ; make new frame pointer
             les     di, ss:[bx + 8]   ; pointer to set buffer
             movzx   ax, byte ptr ss:[bx + 6] ; number of first element to add
             movzx   bx, byte ptr ss:[bx + 4] ; number of last element to add
             cmp     bx, ax            ; first element > last element ?
             jb      @@1               ; yes, nothing to do

             mov     cx, ax            ; leftmost byte
             mov     dx, -1            ; initial masks
             shr     ax, 3             ; offset in buffer
             and     cx, 7             ; first bit set in leftmost byte
             shl     dl, cl            ; make mask for leftmost byte

             mov     cx, bx            ; rightmost byte
             shr     bx, 3             ; offset in buffer
             and     cx, 7             ; get last bit set in that byte
             lea     si, [di + bx]     ; address of rightmost byte
             add     di, ax            ; address of leftmost byte
             inc     cx                ;
             shl     dh, cl            ;
             not     dh                ;

             mov     cx, si            ; compute distance between
             mov     al, 0ffh          ; | value of fill byte
             sub     cx, di            ; leftmost and rightmost byte
             jz      @@2               ; leftmost and rightmost is same byte

             or      es:[di], dl       ; set leftmost byte
             dec     cx                ; number of bytes to fill in the middle
             inc     di                ; first byte in middle part
             rep     stosb             ; fill bytes in the middle, if any
             or      es:[si], dh       ; set rightmost byte
@@1:         retf    4                 ; leave set buffer pointer on stack

@@2:         and     dl, dh            ; select bits in this byte
             or      es:[di], dl       ; set them
             retf    4                 ; leave set buffer pointer on stack


ZAddRange    ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END

comment ~
             mov     bx, sp            ; make new frame pointer
             les     di, ss:[bx + 8]   ; pointer to set buffer
             mov     al, ss:[bx + 6]   ; number of first element to add
             mov     bl, ss:[bx + 4]   ; number of last element to add
             cmp     bl, al            ; first element > last element ?
             jb      $done             ; yes, nothing to do

             mov     cx, 0ff07h        ; initial mask for leftmost byte
             and     cl, al            ; get first bit set in that byte
             shl     ch, cl            ; make mask for leftmost byte
             mov     dl, ch            ; save mask

             mov     cx, 0ff07h        ; initial mask for rightmost byte
             and     cl, bl            ; get last bit set in that byte
             inc     cx                ;
             shl     ch, cl            ;
             not     ch                ;
             mov     dh, ch            ; save mask

             mov     si, di            ; save buffer start
             mov     cl, 3             ; offset of
             shr     al, cl            ;  leftmost byte
             xor     ah, ah            ;   into buffer
             add     di, ax            ; address of leftmost byte
             mov     al, bl            ; get number of last element
             shr     al, cl            ; offset of rightmost byte into buffer
             add     si, ax            ; address of rightmost byte
             mov     cx, si            ; compute distance between
             sub     cx, di            ;  leftmost and rightmost byte
             jz      $one_byte         ; leftmost and rightmost is same byte

             mov     al, 0ffh          ; value of fill byte
             or      es:[di], dl       ; set leftmost byte
             dec     cx                ; number of bytes to fill in the middle
             inc     di                ; first byte in middle part
             rep     stosb             ; fill bytes in the middle, if any
             or      es:[si], dh       ; set rightmost byte

$done:       ret     4                 ; leave set buffer pointer on stack

$one_byte:   and     dl, dh            ; select bits in this byte
             or      es:[di], dl       ; set them
             ret     4                 ; leave set buffer pointer on stack
~