; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     String Insertion                                *
; *                                                     *
; *     Copyright (C) 1990-1993 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   STRHINS

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Publics

             PUBLIC  SInsert

;-------------------------------------------------------------------------------
; Insert standard proc.: Insert (VAR Obj, Target: STRING; MaxLen, Pos: INTEGER);
;
; Insert inserts string Obj in string Target at index Pos. The length of the
; resulting string Target may not exceed MaxLen. If the insertion point is at
; the end of Target, the routine appends Obj to Target. It does this by copying
; characters from Obj to the end of Target until all characters are copied or
; the maximum length MaxLen of Target has been reached. If Obj is to be inserted
; at the start or in the middle of Target, as indicated by Pos, the routine
; first creates a 'gap' in Target and then fills this gap with characters from
; Obj.
;
; On entry:  [SP+4]  Position
;            [SP+6]  Maximum length of target string
;            [SP+8]  Address of target string
;            [SP+12] Address of object string
;-------------------------------------------------------------------------------

SInsert      PROC    FAR

             push    bp                ; save TURBO-Pascal framepointer
             push    ds                ; save TURBO-Pascal data segment
             mov     bp, sp            ; make new framepointer
             lds     si, [bp + 16]     ; address of object string
             xor     cx, cx            ; clear msb of length of object string
             or      cl, [si]          ; get length of object string
             jz      $end_ins          ; exit, if length of object string zero

             mov     dx, [bp + 8]      ; DX=pos of target str.
             mov     ax, [bp + 10]     ; AX = max. length of target str.
             les     di, [bp + 12]     ; address of target string
             movzx   bx, byte ptr es:[di]; get current length of target string
             dec     dx                ; pos - 1
             cmp     dh, 80h           ; is pos-1 < 0 ?
             sbb     bp, bp            ; BP = FFFF if pos-1 > 0, else BP = 0
             and     dx, bp            ; if pos-1 < 0 then set pos-1 = 0 (pos=1)
             sub     dx, bx            ; pos-1 - targetlength
             sbb     bp, bp            ; BP = FFFF, if result negative, else 0
             and     dx, bp            ; pos-1 - targetlength (0, if res. pos)
             add     dx, bx            ; pos-1, if result neg., else target len.
             mov     bp, ax            ; maximum length of target string
             sub     bp, dx            ; (maximum length - pos + 1) <= 0 ?
             jbe     $end_ins          ; yes, don't insert

             sub     bp, cx            ; maximum length - pos + 1 - objectlength
             cld                       ; select auto-increment
             ja      $make_gap         ; jump, if result positiv

             stosb                     ; new targetlength will be maximum length
             add     cx, bp            ; adjust length of object string
             add     di, dx            ; object string inserted after this char
             jmp     $fill             ; insert object

$make_gap:   mov     dx, cx            ; save length of object string
             add     cx, bx            ; new targetlength = objectlen+targetlen
             sub     cx, ax            ; new targetlength - maximum length
             sbb     bx, bx            ; BX=FFFF, if result negative, else BX=0
             and     cx, bx            ; new targetlength - maximum length
             add     ax, cx            ; new targetlength = Min(TargLen, MaxLen)
             mov     es:[di], al       ; store it
             add     cx, bp            ; new targetlength - objectlength - pos+1
             add     di, ax            ; this will be the last char in target
             xchg    ax, si            ; save offset of adress of object string
             mov     si, di            ; now calculate
             sub     si, dx            ;  the starting position
             push    ds                ; save segment of address of objectstring
             mov     bx, es            ; let both segment register point
             mov     ds, bx            ; to the segment containing targetstring
             std                       ; auto-decrement
             rep     movsb             ; create gap
             mov     di, si            ; object will be inserted after this char
             mov     cx, dx            ; get length of object string
             mov     si, ax            ; get offset of address of object string
             pop     ds                ; get segment of address of object string
             inc     di                ; object string will be inserted here
             cld                       ; auto-increment

$fill:       inc     si                ; first char of object string
             shr     cx, 1             ; number of chars to move odd ?
             jnc     $even_ins         ; no

             movsb                     ; yes, move a single byte

$even_ins:   rep     movsw             ; move the rest

$end_ins:    pop     ds                ; restore TURBO-Pascal data segment
             pop     bp                ; restore TURBO-Pascal frame pointer
             ret     12                ; return and pop parameters

SInsert      ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END