; *******************************************************
; *                                                     *
; *     Turbo Pascal Runtime Library Version 6.0/7.0    *
; *     String Pos Function                             *
; *                                                     *
; *     Copyright (C) 1990-1993 Norbert Juffa           *
; *     Copyright (C) 2002-2003 Robert AH Prins         *
; *                                                     *
; *******************************************************

             TITLE   STRHPOS

             .386

CODE         SEGMENT BYTE PUBLIC USE16
             ASSUME  CS:CODE

; Publics

             PUBLIC  SPos

;-------------------------------------------------------------------------------
; Pos standard function: Pos (VAR Obj, Target: STRING): BYTE
;
; Pos searches the string Target for the first occurence of a string Obj. If Obj
; is found in Target, the index at which the first character of Obj appears in
; Target is returned. If Obj is not found in Target, zero is returned. The
; routine searches the first character of Obj in the first Len(Target)-Len(Obj)
; charcters of Target. Whenever there is a match, the following characters of
; Target are compared to the characters in Obj. If all characters match, Obj has
; been succesfully found in Target.
;
; On entry:  [SP+4] Address of Target
;            [SP+8] Address of Obj
;
; On exit:   AX     index at which Obj first found in Target (0 if not found)
;-------------------------------------------------------------------------------

SPos         PROC    FAR

             cld                       ; auto-increment for string instructions
             push    ds                ; save TURBO-Pascal data segment
             push    bp                ; save TURBO-Pascal's frame pointer
             mov     bp, sp            ; make new frame pointer
             les     di, [bp + 8]      ; address of target string
             lds     si, [bp + 12]     ; address of object string
             mov     ax, [si]          ; AL=length of object, AH=1. char of obj.
             add     si, 2             ; adjust string pointer
             dec     ax                ; length - 1 (changes AH only when AL=0)
             movzx   cx, byte ptr es:[di]; length of target string (CH = 0)
             sub     cl, al            ; targetlength <= objectlength - 1 ?
             jbe     $not_found        ; yes, object not found

             mov     bh, cl            ; save length of target string
             inc     di                ; first char of target string
             xchg    al, ah            ; AL=1. char of object, AH=length of obj.
             mov     bp, si            ; save offset into object

$search:     repnz   scasb             ; scan target string for 1. char of obj.
             jnz     $not_found        ; exit, if not found

             mov     bl, cl            ; save remaining length of target string
             mov     cl, ah            ; number of char to be compared
             mov     dx, di            ; save offset into target string
             repz    cmpsb             ; compare remaining char of target & obj.
             mov     si, bp            ; restore offset into object string
             mov     di, dx            ; restore offset into target string
             mov     cl, bl            ; restore remaining length of target str
             jnz     $search           ; cont. search, if not all chars matched

             sub     bh, bl            ; Pos = targetlength - remain. targetlen.
             mov     ch, bh            ; Pos

$not_found:  pop     bp                ; restore TURBO-Pascal's frame pointer
             pop     ds                ; restore TURBO-Pascal data segment
             movzx   ax, ch            ; Pos
             ret     8                 ; return and pop parameters

SPos         ENDP

             INCLUDE ALIGN.ASM

CODE         ENDS

             END